from schematics import types as t, Model

# Container configuration


class ContainerConfiguration(Model):

    app_project_filepath = t.StringType(required=False, default=None)


# Default container
class Container():

    # Custom fields below
    # ...

    def __init__(self, config: ContainerConfiguration):
        # Default init
        self.config = config

        # Custom init below
        # ...


# Default dynamic container
class DynamicContainer():

    def add_service(self, service_name, factory_func):
        setattr(self, service_name, factory_func)



from schematics import Model, types as t
from schematics.types.serializable import serializable
from schematics.transforms import wholelist, whitelist, blacklist


class DefaultOptions():
    serialize_when_none = False
    roles = {
        'to_object': wholelist(),
        'to_data': wholelist()
    }


class DataObject(Model):

    def map(self, type: type, role: str = 'to_object', **kwargs):
        return type(dict(
            **kwargs,
            **self.to_primitive(role=role)
        ), strict=False)



from typing import List, Any
from schematics import types as t, Model
from schematics.transforms import blacklist, whitelist
from schematics.types.serializable import serializable


class AppDomainModel(Model):

    name = t.StringType()
    description = t.StringType()


class AppValueObject(AppDomainModel):
    pass


class AppEntity(AppDomainModel):

    id = t.StringType(required=True)



# Environment
APP_ENV = 'APP_ENV'
DEFAULT_APP_ENV = 'prod'
PROJECTS_FILE_PATH = 'PROJECTS_FILE_PATH'
DEBUG = 'DEBUG'

# Configuration file
APP_CONFIGURATION_FILE = 'app/app.yml'

# Configuration
CONFIGS = 'configs'
ENDPOINTS = 'endpoints'
ERRORS = 'errors'

# Domain constants
DOMAIN_ROLE_TYPES = [
    'whitelist',
    'blacklist'
]



import os
from schematics import types as t, Model


class FeatureConfiguration(Model):

    class FunctionConfiguration(Model):
        name = t.StringType(required=True)
        function_path = t.StringType(required=True)
        data_mapping = t.StringType()
        params = t.DictType(t.StringType(), default={})
        log_activity = t.BooleanType(default=True)

    def __init__(self, raw_data: dict = None, **kwargs):
        # Account for single-module configuration
        try:
            config = {}
            config['modules'] = [{
                'function_path': raw_data.pop('module_path'),
                'params': raw_data.pop('params', {}),
                'data_mapping': raw_data.pop('data_mapping', None),
                'log_activity': raw_data.pop('log_activity', True)
            }]
            config['header_mapping'] = raw_data.pop(
                'header_mapping', None)
            super().__init__(raw_data=config)
        except KeyError:
            super().__init__(raw_data=raw_data)

    name = t.StringType(required=True)
    use_role = t.StringType()
    data_mapping = t.StringType()
    header_mapping = t.StringType()
    functions = t.ListType(t.ModelType(FunctionConfiguration), default=[])
    log_params = t.DictType(t.StringType(), default={})


class FeatureGroupConfiguration(Model):
    data_mapping = t.StringType()
    features = t.DictType(t.ModelType(FeatureConfiguration), default={})


class AppFeaturesConfiguration(Model):
    groups = t.DictType(t.ModelType(FeatureGroupConfiguration), default={})


class ErrorConfiguration(Model):
    error_code = t.StringType(required=True)
    message = t.DictType(t.StringType(), required=True)
    status_code = t.IntType(default=400, choices=[400, 401, 403, 404])  # Status codes include Bad Request, Unauthorized, Forbidden, and Not Found


class AppConfiguration(Model):
    errors = t.DictType(t.ModelType(ErrorConfiguration), default={})


class InterfaceConfiguration(Model):
    type = t.StringType(required=True, choices=['cli', 'rest_flask'])


class AppConfiguration(Model):
    errors = t.DictType(t.ModelType(ErrorConfiguration), default={})
    features = t.ModelType(AppFeaturesConfiguration)
    interfaces = t.DictType(t.ModelType(InterfaceConfiguration), default=[])


class AppConfigurationReader():

    def __init__(self, app_config_filepath: str):
        self.app_config_filepath = app_config_filepath

    def load_config(self, app_name: str, **kwargs) -> AppConfiguration:
        app_name, kwargs
        pass


def load_app_config_reader(app_config_filepath: str) -> AppConfigurationReader:
    if os.path.splitext(app_config_filepath)[1] in ['.yaml', '.yml']:
        from .yaml import YamlAppConfigurationReader
        return YamlAppConfigurationReader(app_config_filepath)
    elif os.path.splitext(app_config_filepath)[1] == '.json':
        from .json import JsonConfigurationReader
        return JsonConfigurationReader(app_config_filepath)



import yaml

from . import *


class YamlAppConfigurationReader(AppConfigurationReader):

    def __init__(self, app_config_filepath: str):
        super().__init__(app_config_filepath)

    def load_config(self, **kwargs) -> AppConfiguration:
        with open(self.app_config_filepath) as stream:
            app_components = yaml.safe_load(stream)
        return AppConfiguration(app_components, **kwargs)



import json

from . import *


class JsonConfigurationReader(AppConfigurationReader):

    def load_config(self, **kwargs) -> AppConfiguration:
        with open(self.app_config_filepath) as stream:
            app_components = json.load(stream)
        return AppConfiguration(app_components, **kwargs)



from schematics import types as t, Model
from schematics.transforms import whitelist


class Error(Model):

    error_code = t.StringType(required=True)
    error_name = t.StringType(required=True)
    message = t.DictType(t.StringType(), required=True)
    # Status codes include Bad Request, Unauthorized, Forbidden, and Not Found
    status_code = t.IntType(default=400, choices=[400, 401, 403, 404])
    format_args = t.ListType(t.StringType(), default=[])
    include_payload = t.BooleanType(default=True)

    class Options:
        roles = {
            'public': whitelist('error_code', 'message', 'status_code')
        }

    def format_message(self, *args):
        self.format_args = args
        return self


class ErrorManager():

    def __init__(self):
        self.__errors = {}

    def get(self, name: str, default=None):
        try:
            return self.__errors[name]
        except KeyError:
            return default

    def add(self, error: Error):
        try:
            self.__errors[error.error_name] = error
            setattr(self, error.error_name, error)
        except:
            return


class AppError(Exception):

    def __init__(self, error: Error, lang: str = 'en_US'):
        super().__init__(self)
        self.error_code = error.error_code
        self.message = error.message[lang]
        if error.format_args:
            self.message = self.message.format(*error.format_args)
        self.status_code = error.status_code

    def to_dict(self):
        return dict(error_code=self.error_code, message=self.message)



from schematics import types as t, Model
from schematics.exceptions import DataError

from . import activity
from .config import FeatureConfiguration
from .error import *

HEADER_MAPPER_PATH = 'app.interfaces.{}.mappers.header'
DATA_MAPPER_PATH = 'app.interfaces.{}.mappers.command'
SERVICES_MAPPER_PATH = 'app.interfaces.services'


class MessageContext():

    def __init__(self):
        self.headers = {}
        self.data = None
        self.services = None
        self.result = {}
        self.errors = ErrorManager()


class FeatureHandler():
    def __init__(self, feature_config):
        if isinstance(feature_config, FeatureConfiguration):
            self.feature_config = feature_config
        elif isinstance(feature_config, dict):
            self.feature_config = FeatureConfiguration(feature_config)
        self.current_step = 0

    def handle(self, request, app_context, **kwargs):
        from time import time
        from importlib import import_module

        # Create message context.
        context: MessageContext = MessageContext()

        # Pull settings first.
        debug = kwargs.get('debug', False)

        # Add errors first.  It's easier this way...
        context.errors = app_context.errors

        # Begin header mapping process.
        if debug:
            print('Perform header mapping: "mapping": "{}"'.format(
                self.feature_config.header_mapping))

        # Import header mappings module.
        header_module = import_module(
            HEADER_MAPPER_PATH.format(app_context.interface))

        try:
            # Retrieve header mapping function.
            header_mapping_func = getattr(
                header_module, self.feature_config.header_mapping)
        except TypeError:
            # Retrieve default header mapping function if none is specified.
            header_mapping_func = getattr(header_module, 'default')

        # Get header data and add to message context.
        context.headers = header_mapping_func(request, app_context, **kwargs)

        feature_group = kwargs.get('feature_group', None)

        for function in self.feature_config.functions:

            if debug:
                print('Executing function: "function": "{}"'.format(
                    function.to_primitive()))

            context.headers['request_start'] = int(time())

            # Set data mapping and service container for feature function
            data_mapping = function.data_mapping
            if not data_mapping:
                data_mapping = self.feature_config.data_mapping
            if not data_mapping:
                data_mapping = feature_group.data_mapping
            try:
                if debug:
                    print('Perform data mapping: "mapping": "{}"'.format(
                        data_mapping))
                data_mapping = getattr(
                    import_module(DATA_MAPPER_PATH.format(
                        app_context.interface)),
                    data_mapping
                )
                if debug:
                    print('Performing data mapping for following request: "mapping": "{}", "request": "{}", params: "{}"'.format(
                        data_mapping, request, function.params))
                context.data = data_mapping(
                    context, request, app_context, **function.params, **kwargs)
                if debug:
                    print('Data mapping complete: "mapping": "{}", "data": "{}"'.format(
                        data_mapping, context.data.to_primitive()))
                # Request model state validation
                try:
                    context.data.validate()
                except AttributeError:  # In the case where there is no configured data mapping
                    pass
            except TypeError as ex:
                print(ex)
            except DataError as ex:
                raise AppError(
                    app_context.errors.INVALID_REQUEST_DATA.format_message(ex.messages))

            context.services = app_context.container

            # Format function module path.
            module_path = 'app.features.{}'.format(function.function_path)

            # Import function module.
            if debug:
                print('Importing function: {}'.format(module_path))
            handler = import_module(module_path)

            # Execute function handler.
            if debug:
                print('Executing function: {}'.format(module_path))
            result = handler.handle(context)
            # For those who do now wish to assign the results to the context in the handler
            if result:
                context.result = result

            # Log activity
            if function.log_activity:
                if debug:
                    print('Logging activity for function: {}'.format(module_path))
                activity.handle(context)

            if debug:
                print('Finishing function: {}'.format(module_path))

        context.headers['request_end'] = int(time())

        # Return result
        # Handle list scenario
        if type(context.result) == list:
            result = []
            for item in context.result:
                if isinstance(item, Model):
                    if self.feature_config.use_role:
                        result.append(item.to_primitive(
                            role=self.feature_config.use_role))
                    else:
                        result.append(item.to_primitive())
                else:
                    result.append(item)
            return result
        if not context.result:
            return {}
        # Convert schematics models to primitive dicts.
        if isinstance(context.result, Model):
            if self.feature_config.use_role:
                return context.result.to_primitive(role=self.feature_config.use_role)
            else:
                return context.result.to_primitive()
        return context.result



try:
    from setuptools import setup
except:
    from distutils.core import setup

config = {
    'description': 'The Core Library for the Spirit of Harmony',
    'author': 'Andrew Shatz',
    'url': r'https://github.com/greatstrength/aikicore',
    'download_url': r'https://github.com/greatstrength/aikicore',
    'author_email': 'andrew@greatstrength.me',
    'version': '0.2.3',
    'license': 'BSD 3',
    'install_requires': [
        'schematics>=2.1.1',
        'pyyaml>=6.0.1'
    ],
    'packages': [
        'aikicore',
        'aikicore.config',
        'aikicore.objects',
    ],    
    'scripts': [],
    'name': 'aikicore'
}

setup(**config)


from typing import Dict

from .constants import APP_CONFIGURATION_FILE
from .config import load_app_config_reader, AppConfigurationReader, AppConfiguration
from .error import *
from .containers import *
from .routing import *


class AppContext():

    name: str = None
    interface: str = None
    container_config: ContainerConfiguration = None
    container: Container = None
    errors: ErrorManager = ErrorManager()
    feature_groups: dict = None

    def __init__(self, name: str, interface: str, app_config: AppConfiguration, container_config: ContainerConfiguration, container: type = Container):
        # Set app name.
        self.name = name

        # Set interface.
        self.interface = interface

        # Load app errors.
        try:
            for error_name, error in app_config.errors.items():
                self.errors.add(
                    Error({'error_name': error_name, **error.to_primitive()}))
        except AttributeError:
            pass

        self.feature_groups = app_config.features.groups

        # Load container config and container.
        self.container_config = container_config
        self.container = container(container_config)

    def run(self, **kwargs):
        pass


class AppBuilder():

    class Session():
        def __init__(self, name: str, app_config: AppConfiguration, container_config: ContainerConfiguration):
            self.name = name
            self.app_config = app_config
            self.container_config = container_config

    _current_session = None

    @property
    def app_config(self) -> AppConfiguration:
        return self._current_session.app_config

    def create_new_app(self, name: str, config_file: str = APP_CONFIGURATION_FILE, **kwargs):
        kwargs
        if self._current_session:
            self._current_session = None
        app_config_reader: AppConfigurationReader = load_app_config_reader(
            config_file)
        app_config = app_config_reader.load_config(strict=False)
        self._current_session = self.Session(
            name=name,
            app_config=app_config,
            container_config=None
        )
        return self

    def set_app_config(self, app_config: AppConfiguration):
        self._current_session.app_config = app_config
        return self

    def set_container_config(self, container_config: ContainerConfiguration):
        self._current_session.container_config = container_config
        return self

    def build(self):
        pass
